// ************************************************************************
// ************************ Micha Kurzela @ 2013 *************************
// ************************** Higrometr binarny ***************************
// ************* Fabryczne ustawienia fusebitw w ATtiny2313 **************
// ************************************************************************

#include <avr/io.h>     
#include <util/delay.h>
  
volatile uint8_t odczyt[5]; // wszystkie 40 bitw pobranych z DHT11 rozoonych na 5 bajtw    
volatile uint8_t i, j; // numer odczytywanego, kolejno: bajtu i bitu
volatile uint8_t stan = 0; // 0 - wskazuje wilgotno; 2 - wskazuje temperatur

//-------------------------------------

void zainicjowanie(void){ // rozpoczcie komunikacji z czujnikiem

DDRD |= 0b01000000; // ustawienie PD.6 jako wyjcie 
PORTD |= 0b01000000; // wystawienie stanu wysokiego 
_delay_ms(50); // stan wysoki trzeba utrzyma do dugo, by DHT11 "zapomnia" o poprzedniej transmisji
      
PORTD &= 0b10111111; // ustawienie stanu niskiego
_delay_ms(20); // nota katalogowa zaleca min. 18ms
  
PORTD |= 0b01000000; // ustawienie stanu wysokiego
_delay_us(20); // nota katalogowa zaleca 20-40us, ale DHT11 reaguje bardzo szybko

DDRD &= 0b10111111; // ustawienie PD.6 jako wejcie, rozpoczcie nasuchiwania  
_delay_us(40); // po podaniu stanu wysokiego na 20us, DHT11 zareaguje poprzez podanie stanw wysokiego i niskiego,
// kady na 80us. Ta przerwa pozwala pozby si zakce wywoanych przez DHT11 "odpowiadajcego" stanem niskim
// na stan wysoki wystawiony przez PD.6

while(bit_is_clear(PIND, PIND6)){;}  // najpierw przeczeka stan niski...
while(bit_is_set(PIND, PIND6)){;} // ...a potem wysoki.
// Teraz DHT11 wyle sekwencj 40 bitw.
}

//-------------------------------------

void czytaj(void){ // odczyt 40 bitw przysanych przez czujnik.
   
for(i = 0; i < 5; i++){ // ptla zewntrzna, ktra porcjuje napywajce bity na bajty
odczyt[i] = 0; // najpierw bajt jest wypeniany zerami

for(j = 0; j < 8; j++){ // ptla wewntrzna, odmierzajca poszczeglne bity w bajcie
odczyt[i] *= 2; // bity przychodz od najstarszego, wic posiadane przeba przesuwa w lewo
while(bit_is_clear(PIND, PIND6)){;} // dwa sowa o sposobie przesyu bitw:
// "zero" jest reprezentowane przez 50us stanu niskiego i 26-28us stanu wysokiego
// "jedynka" z kolei, przez 50us stanu niskiego i 70us stanu wysokiego
//zatem najpierw "przeczekujemy" stan niski na linii, a potem jeszcze 40us
_delay_us(40); // tutaj si czeka te 40us

if(bit_is_set(PIND, PIND6)){ odczyt[i] += 1;} //jeeli po 40us od zaniku stanu niskiego, na linii 
// nadal panuje stan wysoki, naley to interpretowa jako nadejcie "jedynki",
// gdyby byo to "zero", wtedy w momencie sprawdzenia byby stan niski
while(bit_is_set(PIND, PIND6)){;} // jeeli trzeba przeczeka stan wysoki - zrb to
// ptla przejdzie do odczytywania kolejnego bitu

}}}

//-------------------------------------

int main(void){

DDRB = 0b11111111; //konfiguracja nieuywanego portu jako wyjcie
PORTB = 0b11111111;

DDRA = 0b11;
PORTA = 0b11;

DDRD = 0b10011111; // PD.6 - DHT11, PD.5 - switch SW1
PORTD = 0b11111111;

while(1){
zainicjowanie();
czytaj();

stan = 0; //domylnie wywietlana jest wilgotno
if(bit_is_clear(PIND, PIND5)){stan = 2;} //jeeli wcinity jest przycisk, wywietla temperatur
// pierwszy pobrany bajt zawiera wilgotno w % RH
// drugi - czci dziesite wilgotnoci, uywany w DHT22
// trzeci - temperatura w *C
// czwarty - czci dziesite temperatury, uywane w DHT22
// pity - suma kontrolna

// prosty algorytm podziau liczby na bity: jeeil liczba jest wiksza ni X, wwczas odejmij od niej X
// pozosta w ten sposb reszt sprawd analogicznie z Y, przy czym X = 2Y
// i tak do wyczerpania liczby
if(odczyt[stan] >= 64){PORTD &= 0b11111110; odczyt[stan] -= 64;} else {PORTD |= 0b00000001;}
if(odczyt[stan] >= 32){PORTD &= 0b11111101; odczyt[stan] -= 32;} else {PORTD |= 0b00000010;}
if(odczyt[stan] >= 16){PORTA &= 0b01; odczyt[stan] -= 16;} else {PORTA |= 0b10;}
if(odczyt[stan] >= 8){PORTA &= 0b10; odczyt[stan] -= 8;} else {PORTA |= 0b01;}
if(odczyt[stan] >= 4){PORTD &= 0b11111011; odczyt[stan] -= 4;} else {PORTD |= 0b00000100;}
if(odczyt[stan] >= 2){PORTD &= 0b11110111; odczyt[stan] -= 2;} else {PORTD |= 0b00001000;}
if(odczyt[stan] >= 1){PORTD &= 0b11101111; odczyt[stan] -= 1;} else {PORTD |= 0b00010000;}

_delay_ms(1500); //odczyty nie powinny nastpowa czciej jak co sekund
}
}